@@ -1,6 +1,6 @@
 /* defs.h -- data types and declarations.
-   Copyright (C) 1990, 91, 92, 93, 94, 2000, 2004, 2005,
-                 2006, 2007, 2008 Free Software Foundation, Inc.
+   Copyright (C) 1990, 1991, 1992, 1993, 1994, 2000, 2004, 2005, 2006,
+                 2007, 2008, 2009 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -48,6 +48,7 @@ Please stop compiling the program now
 #include <stdbool.h>		/* for bool/boolean */
 #include <stdint.h>		/* for uintmax_t */
 #include <sys/stat.h> /* S_ISUID etc. */
+#include <selinux/selinux.h>
 
 
 
@@ -63,6 +64,7 @@ typedef bool boolean;
 #include "timespec.h"
 #include "buildcmd.h"
 #include "quotearg.h"
+#include "sharefile.h"
 
 /* These days we will assume ANSI/ISO C protootypes work on our compiler. */
 #define PARAMS(Args) Args
@@ -197,6 +199,7 @@ struct exec_val
   boolean use_current_dir;      /* If nonzero, don't chdir to start dir */
   boolean close_stdin;		/* If true, close stdin in the child. */
   int dir_fd;			/* The directory to do the exec in. */
+  int last_child_status;	/* Status of the most recent child. */
 };
 
 /* The format string for a -printf or -fprintf is chopped into one or
@@ -300,7 +303,7 @@ struct predicate
   boolean artificial;
 
   /* The raw text of the argument of this predicate. */
-  char *arg_text;
+  const char *arg_text;
 
   /* Information needed by the predicate processor.
      Next to each member are listed the predicates that use it. */
@@ -318,6 +321,7 @@ struct predicate
     struct samefile_file_id samefileid; /* samefile */
     mode_t type;		/* type */
     struct format_val printf_vec; /* printf fprintf fprint ls fls print0 fprint0 print */
+    security_context_t scontext; /* security context */
   } args;
 
   /* The next predicate in the user input sequence,
@@ -462,11 +466,11 @@ PREDICATEFUNCTION pred_used;
 PREDICATEFUNCTION pred_user;
 PREDICATEFUNCTION pred_writable;
 PREDICATEFUNCTION pred_xtype;
+PREDICATEFUNCTION pred_context;
 
 
 
-int launch PARAMS((struct buildcmd_control *ctl,
-		   struct buildcmd_state *buildstate));
+int launch (struct buildcmd_control *ctl, void *usercontext, int argc, char **argv);
 
 
 char *find_pred_name PARAMS((PRED_FUNC pred_func));
@@ -481,15 +485,20 @@ void show_success_rates(const struct predicate *node);
 
 
 /* tree.c */
+boolean matches_start_point(const char * glob, boolean foldcase);
 struct predicate * build_expression_tree PARAMS((int argc, char *argv[], int end_of_leading_options));
 struct predicate * get_eval_tree PARAMS((void));
+struct predicate *get_new_pred_noarg (const struct parser_table *entry);
 struct predicate *get_new_pred PARAMS((const struct parser_table *entry));
-struct predicate *get_new_pred_chk_op PARAMS((const struct parser_table *entry));
+struct predicate *get_new_pred_chk_op PARAMS((const struct parser_table *entry,
+					      const char *arg));
 float  calculate_derived_rates PARAMS((struct predicate *p));
 
 /* util.c */
-struct predicate *insert_primary PARAMS((const struct parser_table *entry));
-struct predicate *insert_primary_withpred PARAMS((const struct parser_table *entry, PRED_FUNC fptr));
+int fd_leak_check_is_enabled (void);
+struct predicate *insert_primary PARAMS((const struct parser_table *entry, const char *arg));
+struct predicate *insert_primary_noarg PARAMS((const struct parser_table *entry));
+struct predicate *insert_primary_withpred PARAMS((const struct parser_table *entry, PRED_FUNC fptr, const char *arg));
 void usage PARAMS((FILE *fp, int status, char *msg));
 extern boolean check_nofollow(void);
 void complete_pending_execs(struct predicate *p);
@@ -604,6 +613,9 @@ struct options
    */
   int regex_options;
 
+  /* function used to get file context */
+  int (*x_getfilecon) (int, const char *, security_context_t *);
+
   /* Optimisation level.  One is the default.
    */
   unsigned short optimisation_level;
@@ -651,6 +663,9 @@ struct state
    * an optimisation.  Set to true if you want to be conservative.
    */
   boolean execdirs_outstanding;
+
+  /* Shared files, opened via the interface in sharefile.h. */
+  sharefile_handle shared_files;
 };
 
 /* finddata.c */
