@@ -1,6 +1,6 @@
 /* find -- search for files in a directory hierarchy
-   Copyright (C) 1990, 91, 92, 93, 94, 2000,
-                 2003, 2004, 2005, 2007, 2008 Free Software Foundation, Inc.
+   Copyright (C) 1990, 1991, 1992, 1993, 1994, 2000, 2003, 2004, 2005,
+                 2007, 2008, 2009 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -37,7 +37,6 @@
 
 #include <sys/stat.h>
 #include <fcntl.h>
-#include <openat.h>
 
 #include "xalloc.h"
 #include "human.h"
@@ -52,6 +51,7 @@
 #include "quotearg.h"
 #include "xgetcwd.h"
 #include "error.h"
+#include "fdleak.h"
 
 #ifdef HAVE_LOCALE_H
 #include <locale.h>
@@ -131,6 +131,17 @@ main (int argc, char **argv)
   program_name = argv[0];
   state.exit_status = 0;
 
+  if (fd_leak_check_is_enabled ())
+    {
+      remember_non_cloexec_fds ();
+    }
+
+  state.shared_files = sharefile_init("w");
+  if (NULL == state.shared_files)
+    {
+      error (1, errno, _("Failed initialise shared-file hash table"));
+    }
+
   /* Set the option defaults before we do the locale
    * initialisation as check_nofollow() needs to be executed in the
    * POSIX locale.
@@ -183,11 +194,11 @@ main (int argc, char **argv)
     }
 
 
-  starting_desc = open (".", O_RDONLY
+  starting_desc = open_cloexec (".", O_RDONLY
 #if defined O_LARGEFILE
-			|O_LARGEFILE
+				|O_LARGEFILE
 #endif
-			);
+				);
   if (0 <= starting_desc && fchdir (starting_desc) != 0)
     {
       close (starting_desc);
@@ -200,6 +211,7 @@ main (int argc, char **argv)
       if (! starting_dir)
 	error (1, errno, _("cannot get current directory"));
     }
+
   set_stat_placeholders(&starting_stat_buf);
   if ((*options.xstat) (".", &starting_stat_buf) != 0)
     error (1, errno, _("cannot stat current directory"));
@@ -293,7 +305,7 @@ init_mounted_dev_list(int mandatory)
   mounted_devices = get_mounted_devices(&num_mounted_devices);
   if (mandatory && (NULL == mounted_devices))
     {
-      error(1, 0, "Cannot read list of mounted devices.");
+      error(1, 0, _("Cannot read list of mounted devices."));
     }
 }
 
@@ -578,11 +590,11 @@ safely_chdir_lstat(const char *dest,
 
   saved_errno = errno = 0;
 
-  dotfd = open(".", O_RDONLY
+  dotfd = open_cloexec (".", O_RDONLY
 #if defined O_LARGEFILE
-	       |O_LARGEFILE
+			|O_LARGEFILE
 #endif
-	       );
+			);
 
   /* We jump back to here if wd_sanity_check()
    * recoverably triggers an alert.
@@ -690,7 +702,7 @@ safely_chdir_lstat(const char *dest,
 			   * is fatal.
 			   */
 			  error(1, errno,
-				"failed to return to parent directory");
+				_("failed to return to parent directory"));
 			}
 		    }
 		  else
@@ -819,6 +831,9 @@ safely_chdir_nofollow(const char *dest,
   fd = open(dest, O_RDONLY
 #if defined O_LARGEFILE
 	    |O_LARGEFILE
+#endif
+#if defined O_CLOEXEC
+	    |O_CLOEXEC
 #endif
 	    |extraflags);
   if (fd < 0)
