@@ -197,10 +197,10 @@ must_read_fs_list(bool need_fs_type)
   if (NULL == entries)
     {
       /* We cannot determine for sure which file we were trying to
-       * use because gnulib has extracted all that stuff away.
+       * use because gnulib has abstracted all that stuff away.
        * Hence we cannot issue a specific error message here.
        */
-      error(1, 0, "Cannot read mounted file system list");
+      error(1, 0, _("Cannot read mounted file system list"));
     }
   return entries;
 }
