@@ -49,11 +49,12 @@
 #include "quotearg.h"
 #include "quote.h"
 #include "fts_.h"
-#include "openat.h"
 #include "save-cwd.h"
 #include "xgetcwd.h"
 #include "error.h"
 #include "dircallback.h"
+#include "cloexec.h"
+#include "fdleak.h"
 
 #ifdef HAVE_LOCALE_H
 #include <locale.h>
@@ -75,21 +76,6 @@
 #endif
 
 
-static void set_close_on_exec(int fd)
-{
-#if defined F_GETFD && defined FD_CLOEXEC
-  int flags;
-  flags = fcntl(fd, F_GETFD);
-  if (flags >= 0)
-    {
-      flags |= FD_CLOEXEC;
-      fcntl(fd, F_SETFD, flags);
-    }
-#endif
-}
-
-
-
 /* FTS_TIGHT_CYCLE_CHECK tries to work around Savannah bug #17877
  * (but actually using it doesn't fix the bug).
  */
@@ -152,8 +138,7 @@ static void inside_dir(int dir_fd)
 	    }
 	  else if (dir_fd >= 0)
 	    {
-	      curr_fd = dup(dir_fd);
-	      set_close_on_exec(curr_fd);
+	      curr_fd = dup_cloexec (dir_fd);
 	    }
 	  else
 	    {
@@ -587,13 +572,12 @@ consider_visiting(FTS *p, FTSENT *ent)
 
 
 static void
-find(char *arg)
+find (char *arg)
 {
   char * arglist[2];
   FTS *p;
   FTSENT *ent;
 
-
   state.starting_path_length = strlen(arg);
   inside_dir(AT_FDCWD);
 
@@ -618,7 +602,7 @@ find(char *arg)
   if (options.stay_on_filesystem)
     ftsoptions |= FTS_XDEV;
 
-  p = fts_open(arglist, ftsoptions, NULL);
+  p = fts_open (arglist, ftsoptions, NULL);
   if (NULL == p)
     {
       error (0, errno, _("cannot search %s"),
@@ -648,7 +632,7 @@ process_all_startpoints(int argc, char *argv[])
   for (i = 0; i < argc && !looks_like_expression(argv[i], true); i++)
     {
       state.starting_path_length = strlen(argv[i]); /* TODO: is this redundant? */
-      find(argv[i]);
+      find (argv[i]);
     }
 
   if (i == 0)
@@ -678,6 +662,17 @@ main (int argc, char **argv)
   state.execdirs_outstanding = false;
   state.cwd_dir_fd = AT_FDCWD;
 
+  if (fd_leak_check_is_enabled ())
+    {
+      remember_non_cloexec_fds ();
+    }
+
+  state.shared_files = sharefile_init("w");
+  if (NULL == state.shared_files)
+    {
+      error (1, errno, _("Failed initialise shared-file hash table"));
+    }
+
   /* Set the option defaults before we do the locale initialisation as
    * check_nofollow() needs to be executed in the POSIX locale.
    */
@@ -727,11 +722,11 @@ main (int argc, char **argv)
     }
 
 
-  starting_desc = open (".", O_RDONLY
+  starting_desc = open_cloexec (".", O_RDONLY
 #if defined O_LARGEFILE
-			|O_LARGEFILE
+				|O_LARGEFILE
 #endif
-			);
+				);
   if (0 <= starting_desc && fchdir (starting_desc) != 0)
     {
       close (starting_desc);
@@ -744,7 +739,6 @@ main (int argc, char **argv)
 	error (1, errno, _("cannot get current directory"));
     }
 
-
   process_all_startpoints(argc-end_of_leading_options, argv+end_of_leading_options);
 
   /* If "-exec ... {} +" has been used, there may be some
