@@ -31,8 +31,8 @@
 #include <stdarg.h>
 #include <fcntl.h>
 #include <locale.h>
-#include <openat.h>
 #include <ctype.h>
+#include <unistd.h> /* for unlinkat() */
 #include "xalloc.h"
 #include "dirname.h"
 #include "human.h"
@@ -46,6 +46,9 @@
 #include "dircallback.h"
 #include "error.h"
 #include "verify.h"
+#include "fdleak.h"
+
+#include <selinux/selinux.h>
 
 #if ENABLE_NLS
 # include <libintl.h>
@@ -230,6 +233,7 @@ struct pred_assoc pred_table[] =
   {pred_user, "user    "},
   {pred_writable, "writable "},
   {pred_xtype, "xtype   "},
+  {pred_context, "context"},
   {0, "none    "}
 };
 #endif
@@ -546,8 +550,18 @@ new_impl_pred_exec (int dir_fd, const char *pathname,
 	}
 
       /* Actually invoke the command. */
-      return  execp->ctl.exec_callback(&execp->ctl,
-					&execp->state);
+      bc_do_exec (&execp->ctl, &execp->state);
+      if (WIFEXITED(execp->last_child_status))
+	{
+	  if (0 == WEXITSTATUS(execp->last_child_status))
+	    return true;	/* The child succeeded. */
+	  else
+	    return false;
+	}
+      else
+	{
+	  return false;
+	}
     }
 }
 
@@ -1054,6 +1068,30 @@ do_fprintf(struct format_val *dest,
 			     mode_to_filetype(stat_buf->st_mode & S_IFMT));
 	  }
 	  break;
+
+	case 'Z':               /* SELinux security context */
+	  {
+	    security_context_t scontext;
+	    int rv = (*options.x_getfilecon) (state.cwd_dir_fd, state.rel_pathname,
+					      &scontext);
+	    if (rv < 0)
+	      {
+		/* If getfilecon fails, there will in the general case
+		   still be some text to print.   We just make %Z expand
+		   to an empty string. */
+		checked_fprintf (dest, segment->text, "");
+
+		error (0, errno, "getfilecon: %s",
+		    safely_quote_err_filename (0, pathname));
+		state.exit_status = 1;
+	      }
+	    else
+	      {
+		checked_fprintf (dest, segment->text, scontext);
+		freecon (scontext);
+	      }
+	  }
+	  break;
 	}
       /* end of KIND_FORMAT case */
       break;
@@ -1866,6 +1904,26 @@ pred_xtype (const char *pathname, struct stat *stat_buf, struct predicate *pred_
    */
   return (pred_type (pathname, &sbuf, pred_ptr));
 }
+
+
+boolean
+pred_context (const char *pathname, struct stat *stat_buf,
+	      struct predicate *pred_ptr)
+{
+  security_context_t scontext;
+  int rv = (*options.x_getfilecon) (state.cwd_dir_fd, state.rel_pathname,
+				    &scontext);
+  if (rv < 0)
+    {
+      error (0, errno, "getfilecon: %s", safely_quote_err_filename (0, pathname));
+      return false;
+    }
+
+  rv = (fnmatch (pred_ptr->args.scontext, scontext, 0) == 0);
+  freecon (scontext);
+  return rv;
+}
+
 
 /*  1) fork to get a child; parent remembers the child pid
     2) child execs the command requested
@@ -1940,14 +1998,15 @@ prep_child_for_exec (boolean close_stdin, int dir_fd)
 
 
 
+
+
+
 int
-launch (struct buildcmd_control *ctl,
-	struct buildcmd_state *buildstate)
+launch (struct buildcmd_control *ctl, void *usercontext, int argc, char **argv)
 {
-  int wait_status;
   pid_t child_pid;
   static int first_time = 1;
-  const struct exec_val *execp = buildstate->usercontext;
+  struct exec_val *execp = usercontext;
 
   if (!execp->use_current_dir)
     {
@@ -1955,10 +2014,6 @@ launch (struct buildcmd_control *ctl,
       assert (execp->dir_fd == starting_desc);
     }
 
-
-  /* Null terminate the arg list.  */
-  bc_push_arg (ctl, buildstate, (char *) NULL, 0, NULL, 0, false);
-
   /* Make sure output of command doesn't get mixed with find output. */
   fflush (stdout);
   fflush (stderr);
@@ -1981,35 +2036,40 @@ launch (struct buildcmd_control *ctl,
 	{
 	  _exit(1);
 	}
+      else
+	{
+	  if (fd_leak_check_is_enabled ())
+	    {
+	      complain_about_leaky_fds ();
+	    }
+	}
 
-      execvp (buildstate->cmd_argv[0], buildstate->cmd_argv);
+      if (bc_args_exceed_testing_limit (argv))
+	errno = E2BIG;
+      else
+	execvp (argv[0], argv);
+      /* TODO: use a pipe to pass back the errno value, like xargs does */
       error (0, errno, "%s",
-	     safely_quote_err_filename(0, buildstate->cmd_argv[0]));
+	     safely_quote_err_filename(0, argv[0]));
       _exit (1);
     }
 
-
-  /* In parent; set up for next time. */
-  bc_clear_args(ctl, buildstate);
-
-
-  while (waitpid (child_pid, &wait_status, 0) == (pid_t) -1)
+  while (waitpid (child_pid, &(execp->last_child_status), 0) == (pid_t) -1)
     {
       if (errno != EINTR)
 	{
 	  error (0, errno, _("error waiting for %s"),
-		 safely_quote_err_filename(0, buildstate->cmd_argv[0]));
+		 safely_quote_err_filename(0, argv[0]));
 	  state.exit_status = 1;
 	  return 0;		/* FAIL */
 	}
     }
 
-  if (WIFSIGNALED (wait_status))
+  if (WIFSIGNALED (execp->last_child_status))
     {
       error (0, 0, _("%s terminated by signal %d"),
-	     quotearg_n_style(0, options.err_quoting_style,
-			      buildstate->cmd_argv[0]),
-	     WTERMSIG (wait_status));
+	     quotearg_n_style(0, options.err_quoting_style, argv[0]),
+	     WTERMSIG (execp->last_child_status));
 
       if (execp->multiple)
 	{
@@ -2023,7 +2083,7 @@ launch (struct buildcmd_control *ctl,
       return 1;			/* OK */
     }
 
-  if (0 == WEXITSTATUS (wait_status))
+  if (0 == WEXITSTATUS (execp->last_child_status))
     {
       return 1;			/* OK */
     }
@@ -2037,7 +2097,10 @@ launch (struct buildcmd_control *ctl,
 	   */
 	  state.exit_status = 1;
 	}
-      return 0;			/* FAIL */
+      /* The child failed, but this is the exec callback.  We
+       * don't want to run the child again in this case anwyay.
+       */
+      return 1;			/* FAIL (but don't try again) */
     }
 
 }
