@@ -20,6 +20,7 @@
 
 #include <assert.h>
 #include <stdlib.h>
+#include <fnmatch.h>
 
 #include "xalloc.h"
 #include "error.h"
@@ -49,6 +50,11 @@ static struct predicate *eval_tree  = NULL;
 /* The last predicate allocated. */
 static struct predicate *last_pred = NULL;
 
+/* The starting points. */
+static char **start_points;
+static size_t num_start_points = 0;
+
+
 
 static struct predicate *scan_rest PARAMS((struct predicate **input,
 				       struct predicate *head,
@@ -58,6 +64,34 @@ static struct predicate *set_new_parent PARAMS((struct predicate *curr, enum pre
 static const char *cost_name PARAMS((enum EvaluationCost cost));
 
 
+/* Return true if the indicated path name is a start
+   point or not.   If no start points were given on the
+   command line, we return true for ".".
+*/
+boolean
+matches_start_point(const char *glob, bool foldcase)
+{
+  int fnmatch_flags = 0;
+  if (foldcase)
+    fnmatch_flags |= FNM_CASEFOLD;
+
+  if (num_start_points)
+    {
+      size_t i;
+      for (i=0; i<num_start_points; i++)
+	{
+	  if (fnmatch (glob, start_points[i], fnmatch_flags) == 0)
+	    return true;
+	}
+      return false;
+    }
+  else
+    {
+      return fnmatch (glob, ".", fnmatch_flags) == 0;
+    }
+}
+
+
 /* Return a pointer to a tree that represents the
    expression prior to non-unary operator *INPUT.
    Set *INPUT to point at the next input predicate node.
@@ -269,10 +303,14 @@ predicate_is_cost_free(const struct predicate *p)
 /* Prints a predicate */
 void print_predicate(FILE *fp, const struct predicate *p)
 {
-  fprintf (fp, "%s%s%s",
-	   p->p_name,
-	   p->arg_text ? " " : "",
-	   p->arg_text ? p->arg_text : "");
+  if (p->arg_text)
+    {
+      fprintf (fp, "%s %s", p->p_name, p->arg_text);
+    }
+  else
+    {
+      fprintf (fp, "%s", p->p_name);
+    }
 }
 
 
@@ -834,6 +872,7 @@ set_new_parent (struct predicate *curr, enum predicate_precedence high_prec, str
   new_parent->need_type = false;
   new_parent->need_inum = false;
   new_parent->p_cost = NeedsNothing;
+  new_parent->arg_text = NULL;
 
   switch (high_prec)
     {
@@ -906,6 +945,7 @@ static struct pred_cost_lookup costlookup[] =
     { pred_cmin      ,  NeedsStatInfo,       },
     { pred_cnewer    ,  NeedsStatInfo,       },
     { pred_comma     ,  NeedsNothing,        },
+    { pred_context   ,  NeedsAccessInfo      },
     { pred_ctime     ,  NeedsStatInfo,       },
     { pred_delete    ,  NeedsSyncDiskHit     },
     { pred_empty     ,  NeedsStatInfo        },
@@ -1204,11 +1244,13 @@ build_expression_tree(int argc, char *argv[], int end_of_leading_options)
   predicates = NULL;
 
   /* Find where in ARGV the predicates begin by skipping the list of
-   * start points.
+   * start points.  As a side effect, also figure out which is the
+   * first and last start point.
    */
+  start_points = argv + end_of_leading_options;
   for (i = end_of_leading_options; i < argc && !looks_like_expression(argv[i], true); i++)
     {
-      /* Do nothing. */ ;
+      ++num_start_points;
     }
 
   /* Enclose the expression in `( ... )' so a default -print will
@@ -1399,6 +1441,18 @@ init_pred_perf(struct predicate *pred)
   p->visits = p->successes = 0;
 }
 
+
+struct predicate *
+get_new_pred_noarg (const struct parser_table *entry)
+{
+  struct predicate *p = get_new_pred(entry);
+  if (p)
+    {
+      p->arg_text = NULL;
+    }
+  return p;
+}
+
 
 /* Return a pointer to a new predicate structure, which has been
    linked in as the last one in the predicates list.
@@ -1440,7 +1494,10 @@ get_new_pred (const struct parser_table *entry)
   last_pred->need_stat = true;
   last_pred->need_type = true;
   last_pred->need_inum = false;
+  last_pred->p_cost = NeedsUnknown;
+  last_pred->arg_text = "ThisShouldBeSetToSomethingElse";
   last_pred->args.str = NULL;
+  last_pred->args.scontext = NULL;
   last_pred->pred_next = NULL;
   last_pred->pred_left = NULL;
   last_pred->pred_right = NULL;
@@ -1456,7 +1513,8 @@ get_new_pred (const struct parser_table *entry)
    predicate is an operator.  If it isn't, the AND operator is inserted. */
 
 struct predicate *
-get_new_pred_chk_op (const struct parser_table *entry)
+get_new_pred_chk_op (const struct parser_table *entry,
+		     const char *arg)
 {
   struct predicate *new_pred;
   static const struct parser_table *entry_and = NULL;
@@ -1478,7 +1536,7 @@ get_new_pred_chk_op (const struct parser_table *entry)
       case PRIMARY_TYPE:
       case CLOSE_PAREN:
 	/* We need to interpose the and operator. */
-	new_pred = get_new_pred (entry_and);
+	new_pred = get_new_pred_noarg (entry_and);
 	new_pred->pred_func = pred_and;
 	new_pred->p_name = "-a";
 	new_pred->p_type = BI_OP;
@@ -1486,6 +1544,7 @@ get_new_pred_chk_op (const struct parser_table *entry)
 	new_pred->need_stat = false;
 	new_pred->need_type = false;
 	new_pred->need_inum = false;
+	new_pred->arg_text = NULL;
 	new_pred->args.str = NULL;
 	new_pred->side_effects = false;
 	new_pred->no_default_print = false;
@@ -1496,6 +1555,7 @@ get_new_pred_chk_op (const struct parser_table *entry)
       }
 
   new_pred = get_new_pred (entry);
+  new_pred->arg_text = arg;
   new_pred->parser_entry = entry;
   return new_pred;
 }
