@@ -1,6 +1,6 @@
 /* util.c -- functions for initializing new tree elements, and other things.
    Copyright (C) 1990, 91, 92, 93, 94, 2000, 2003, 2004, 2005,
-                 2008 Free Software Foundation, Inc.
+                 2008, 2009 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -24,6 +24,7 @@
 #include <sys/utsname.h>
 #endif
 #include <sys/time.h>
+#include <sys/stat.h> /* for fstatat() */
 #include <ctype.h>
 #include <string.h>
 #include <limits.h>
@@ -35,7 +36,7 @@
 #include "timespec.h"
 #include "error.h"
 #include "verify.h"
-#include "openat.h"
+
 
 #if ENABLE_NLS
 # include <libintl.h>
@@ -89,11 +90,13 @@ static struct debug_option_assoc debugassoc[] =
    operator. */
 
 struct predicate *
-insert_primary_withpred (const struct parser_table *entry, PRED_FUNC pred_func)
+insert_primary_withpred (const struct parser_table *entry,
+			 PRED_FUNC pred_func,
+			 const char *arg)
 {
   struct predicate *new_pred;
 
-  new_pred = get_new_pred_chk_op (entry);
+  new_pred = get_new_pred_chk_op (entry, arg);
   new_pred->pred_func = pred_func;
   new_pred->p_name = entry->parser_name;
   new_pred->args.str = NULL;
@@ -118,10 +121,16 @@ insert_primary_withpred (const struct parser_table *entry, PRED_FUNC pred_func)
    either not there at all (we are the very first node) or is an
    operator. */
 struct predicate *
-insert_primary (const struct parser_table *entry)
+insert_primary (const struct parser_table *entry, const char *arg)
 {
   assert (entry->pred_func != NULL);
-  return insert_primary_withpred(entry, entry->pred_func);
+  return insert_primary_withpred(entry, entry->pred_func, arg);
+}
+
+struct predicate *
+insert_primary_noarg (const struct parser_table *entry)
+{
+  return insert_primary(entry, NULL);
 }
 
 
@@ -352,7 +361,7 @@ do_complete_pending_execdirs(struct predicate *p, int dir_fd)
 	  if (execp->state.todo)
 	    {
 	      /* There are not-yet-executed arguments. */
-	      launch (&execp->ctl, &execp->state);
+	      bc_do_exec (&execp->ctl, &execp->state);
 	    }
 	}
     }
@@ -400,7 +409,7 @@ complete_pending_execs(struct predicate *p)
       if (execp->state.todo)
 	{
 	  /* There are not-yet-executed arguments. */
-	  launch (&execp->ctl, &execp->state);
+	  bc_do_exec (&execp->ctl, &execp->state);
 	}
     }
 
@@ -419,55 +428,58 @@ traverse_tree(struct predicate *tree,
   if (tree->pred_right)
     traverse_tree(tree->pred_right, callback);
 }
-
+
+/* After sharefile_destroy is called, our output file
+ * pointers will be dangling (fclose will already have
+ * been called on them).  NULL these out.
+ */
 static void
-flush_and_close_output_files(struct predicate *p)
+undangle_file_pointers (struct predicate *p)
 {
-  if (pred_is(p, pred_fprint)
-      || pred_is(p, pred_fprintf)
-      || pred_is(p, pred_fls)
-      || pred_is(p, pred_fprint0))
+  if (pred_is (p, pred_fprint)
+      || pred_is (p, pred_fprintf)
+      || pred_is (p, pred_fls)
+      || pred_is (p, pred_fprint0))
     {
-      FILE *f = p->args.printf_vec.stream;
-      bool failed;
+      /* The file was already fclose()d by sharefile_destroy. */
+      p->args.printf_vec.stream = NULL;
+    }
+}
 
-      if (f == stdout || f == stderr)
-	failed = fflush(p->args.printf_vec.stream) == EOF;
-      else
-	failed = fclose(p->args.printf_vec.stream) == EOF;
-
-      if (failed)
-	  nonfatal_file_error(p->args.printf_vec.filename);
-    }
-  else if (pred_is(p, pred_print))
-    {
-      if (fflush(p->args.printf_vec.stream) == EOF)
-	{
-	  nonfatal_file_error(p->args.printf_vec.filename);
-	}
-    }
-  else if (pred_is(p, pred_ls) || pred_is(p, pred_print0))
-    {
-      if (fflush(stdout) == EOF)
-	{
-	  /* XXX: migrate to printf_vec. */
-	  nonfatal_file_error("standard output");
-	}
-    }
+/* Return nonzero if file descriptor leak-checking is enabled.
+ */
+int
+fd_leak_check_is_enabled (void)
+{
+  return getenv ("GNU_FINDUTILS_FD_LEAK_CHECK");
 }
 
 /* Complete any outstanding commands.
+ * Flush and close any open files.
  */
 void
-cleanup(void)
+cleanup (void)
 {
   struct predicate *eval_tree = get_eval_tree();
   if (eval_tree)
     {
       traverse_tree(eval_tree, complete_pending_execs);
       complete_pending_execdirs(get_current_dirfd());
-      traverse_tree(eval_tree, flush_and_close_output_files);
     }
+
+  /* Close ouptut files and NULL out references to them. */
+  sharefile_destroy (state.shared_files);
+  if (eval_tree)
+    traverse_tree(eval_tree, undangle_file_pointers);
+
+  if (fd_leak_check_is_enabled ())
+    {
+      complain_about_leaky_fds ();
+      forget_non_cloexec_fds ();
+    }
+
+  if (fflush (stdout) == EOF)
+    nonfatal_file_error ("standard output");
 }
 
 /* Savannah bug #16378 manifests as an assertion failure in pred_type()
@@ -802,6 +814,7 @@ process_debug_options(char *arg)
       exit(0);
     }
 }
+
 
 static void
 process_optimisation_option(const char *arg)
