@@ -236,7 +236,7 @@ static int read_line PARAMS ((void));
 static int read_string PARAMS ((void));
 static boolean print_args PARAMS ((boolean ask));
 /* static void do_exec PARAMS ((void)); */
-static int xargs_do_exec (struct buildcmd_control *cl, struct buildcmd_state *state);
+static int xargs_do_exec (struct buildcmd_control *ctl, void *usercontext, int argc, char **argv);
 static void exec_if_possible PARAMS ((void));
 static void add_proc PARAMS ((pid_t pid));
 static void wait_for_proc PARAMS ((boolean all, unsigned int minreap));
@@ -370,6 +370,16 @@ fail_due_to_env_size (void)
   error (1, 0, _("environment is too large for exec"));
 }
 
+static size_t
+smaller_of (size_t a, size_t b)
+{
+  if (a < b)
+    return a;
+  else
+    return b;
+}
+
+
 
 int
 main (int argc, char **argv)
@@ -438,27 +448,30 @@ main (int argc, char **argv)
        * specifies that it shall be at least LINE_MAX.
        */
       long val;
-#ifdef ARG_MAX
-      assert(bc_ctl.arg_max <= (ARG_MAX-2048));
-#endif
 #ifdef _SC_ARG_MAX
-      val = sysconf(_SC_ARG_MAX);
+      val = sysconf (_SC_ARG_MAX);
       if (val > 0)
 	{
+	  assert (val > XARGS_POSIX_HEADROOM);
 	  /* Note that val can in fact be greater than ARG_MAX
 	   * and bc_ctl.arg_max can also be greater than ARG_MAX.
 	   */
-	  assert (bc_ctl.arg_max <= (val-XARGS_POSIX_HEADROOM));
+	  bc_ctl.arg_max = smaller_of (bc_ctl.arg_max,
+				       (size_t)val-XARGS_POSIX_HEADROOM);
 	}
       else
 	{
 # if defined ARG_MAX
-	  assert (bc_ctl.arg_max <= (ARG_MAX-XARGS_POSIX_HEADROOM));
+	  assert (ARG_MAX > XARGS_POSIX_HEADROOM);
+	  bc_ctl.arg_max = smaller_of (bc_ctl.arg_max,
+				       (ARG_MAX - XARGS_POSIX_HEADROOM));
 # endif
 	}
 #else
       /* No _SC_ARG_MAX */
-      assert (bc_ctl.arg_max <= (ARG_MAX-XARGS_POSIX_HEADROOM));
+      assert (ARG_MAX > XARGS_POSIX_HEADROOM);
+      bc_ctl.arg_max = smaller_of (bc_ctl.arg_max,
+				   (ARG_MAX - XARGS_POSIX_HEADROOM));
 #endif
 
 
@@ -1006,7 +1019,7 @@ print_args (boolean ask)
 	{
 	  tty_stream = fopen ("/dev/tty", "r");
 	  if (!tty_stream)
-	    error (1, errno, "/dev/tty");
+	    error (1, errno, _("failed to open /dev/tty for reading"));
 	}
       fputs ("?...", stderr);
       fflush (stderr);
@@ -1057,7 +1070,7 @@ prep_child_for_exec (void)
    as an atexit() function.
 */
 static int
-xargs_do_exec (struct buildcmd_control *ctl, struct buildcmd_state *state)
+xargs_do_exec (struct buildcmd_control *ctl, void *usercontext, int argc, char **argv)
 {
   pid_t child;
   int fd[2];
@@ -1065,7 +1078,6 @@ xargs_do_exec (struct buildcmd_control *ctl, struct buildcmd_state *state)
   int r;
 
   (void) ctl;
-  (void) state;
 
   if (!query_before_executing || print_args (true))
     {
@@ -1090,7 +1102,7 @@ xargs_do_exec (struct buildcmd_control *ctl, struct buildcmd_state *state)
       wait_for_proc (false, 0u);
 
       if (pipe(fd))
-	error (1, errno, "could not create pipe before fork");
+	error (1, errno, _("could not create pipe before fork"));
       fcntl(fd[1], F_SETFD, FD_CLOEXEC);
 
       /* If we run out of processes, wait for a child to return and
@@ -1110,7 +1122,10 @@ xargs_do_exec (struct buildcmd_control *ctl, struct buildcmd_state *state)
 
 	    prep_child_for_exec();
 
-	    execvp (bc_state.cmd_argv[0], bc_state.cmd_argv);
+	    if (bc_args_exceed_testing_limit (argv))
+	      errno = E2BIG;
+	    else
+	      execvp (argv[0], argv);
 	    if (errno)
 	      {
 		/* Write errno value to parent.  We do this even if
@@ -1131,13 +1146,13 @@ xargs_do_exec (struct buildcmd_control *ctl, struct buildcmd_state *state)
 	    close(fd[1]);
 	    if (E2BIG != errno)
 	      {
-		error (0, errno, "%s", bc_state.cmd_argv[0]);
-		/* The actual value returned here should be irrelevant,
-		 * because the parent will test our value of errno.
-		 */
-		_exit (errno == ENOENT ? 127 : 126);
-
+		error (0, errno, "%s", argv[0]);
 	      }
+	    /* The actual value returned here should be irrelevant,
+	     * because the parent will test our value of errno.
+	     */
+	    _exit (errno == ENOENT ? 127 : 126);
+
 	  /*NOTREACHED*/
 	  } /* child */
 
